use std::fmt::Write;
use std::net::{Ipv4Addr, Ipv6Addr};

use domain::base::rdata::ComposeRecordData;
use domain::base::scan::Symbol;
use domain::base::wire::{Composer, ParseError};
use domain::base::{Name, ParseRecordData, ParsedName, RecordData, Rtype, ToName, Ttl};
use domain::rdata;
use domain::dep::octseq::{Parser, Octets};
use serde::{Deserialize, Serialize};

use crate::errors::Error;
use crate::validation;

use crate::macros::{append_errors, push_error};
use super::record::RecordParseError;

/// Type used to serialize / deserialize resource records data to response / request
///
#[derive(Debug, Deserialize, Serialize)]
#[serde(tag = "type", content = "rdata")]
#[serde(rename_all = "UPPERCASE")]
pub enum RData {
    A(A),
    Aaaa(Aaaa),
    // TODO: CAA
    Cname(Cname),
    // TODO: DS
    Mx(Mx),
    Ns(Ns),
    Ptr(Ptr),
    Soa(Soa),
    Srv(Srv),
    // TODO: SSHFP
    // TODO: SVCB / HTTPS
    // TODO: TLSA
    Txt(Txt),
}

pub enum ParsedRData<Name, Octs> {
    A(rdata::A),
    Aaaa(rdata::Aaaa),
    Cname(rdata::Cname<Name>),
    Mx(rdata::Mx<Name>),
    Ns(rdata::Ns<Name>),
    Ptr(rdata::Ptr<Name>),
    Soa(rdata::Soa<Name>),
    Srv(rdata::Srv<Name>),
    Txt(rdata::Txt<Octs>),
}

impl<Name: ToString, Octs: AsRef<[u8]>> From<ParsedRData<Name, Octs>> for RData {
    fn from(value: ParsedRData<Name, Octs>) -> Self {
        match value {
            ParsedRData::A(record_rdata) => RData::A(record_rdata.into()),
            ParsedRData::Aaaa(record_rdata) => RData::Aaaa(record_rdata.into()),
            ParsedRData::Cname(record_rdata) => RData::Cname(record_rdata.into()),
            ParsedRData::Mx(record_rdata) => RData::Mx(record_rdata.into()),
            ParsedRData::Ns(record_rdata) => RData::Ns(record_rdata.into()),
            ParsedRData::Ptr(record_rdata) => RData::Ptr(record_rdata.into()),
            ParsedRData::Soa(record_rdata) => RData::Soa(record_rdata.into()),
            ParsedRData::Srv(record_rdata) => RData::Srv(record_rdata.into()),
            ParsedRData::Txt(record_rdata) => RData::Txt(record_rdata.into()),
        }
    }
}

impl TryFrom<RData> for ParsedRData<Name<Vec<u8>>, Vec<u8>> {
    type Error = Vec<Error>;

    fn try_from(value: RData) -> Result<Self, Self::Error> {
        let rdata = match value {
            RData::A(record_rdata) => ParsedRData::A(record_rdata.parse_record()?),
            RData::Aaaa(record_rdata) => ParsedRData::Aaaa(record_rdata.parse_record()?),
            RData::Cname(record_rdata) => ParsedRData::Cname(record_rdata.parse_record()?),
            RData::Mx(record_rdata) => ParsedRData::Mx(record_rdata.parse_record()?),
            RData::Ns(record_rdata) => ParsedRData::Ns(record_rdata.parse_record()?),
            RData::Ptr(record_rdata) => ParsedRData::Ptr(record_rdata.parse_record()?),
            RData::Soa(record_rdata) => ParsedRData::Soa(record_rdata.parse_record()?),
            RData::Srv(record_rdata) => ParsedRData::Srv(record_rdata.parse_record()?),
            RData::Txt(record_rdata) => ParsedRData::Txt(record_rdata.parse_record()?),
        };
        Ok(rdata)
    }
}



macro_rules! parse_name {
    ($value:expr, $field:ident, $rtype:literal, $errors:expr) => {
        {
            let name = push_error!(
                validation::normalize_domain(&$value.$field),
                $errors, concat!("/", stringify!($field))
            );

            let name = name.and_then(|name| {
                push_error!(
                    name.parse::<Name<_>>().map_err(|e| {
                        Error::from(RecordParseError::RDataUnknown {
                            input: $value.$field,
                            field: stringify!(field).to_string(),
                            rtype: $rtype.to_string(),
                        }).with_cause(&e.to_string())
                    }),
                    $errors, concat!("/", stringify!($field))
                )
            });

            name
        }
    };
}

/* --------- A --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct A {
    pub address: String,
}

impl From<rdata::A> for A {
    fn from(record_data: rdata::A) -> Self {
        A { address: record_data.addr().to_string() }
    }
}

impl A {
    pub fn parse_record(self) -> Result<rdata::A, Vec<Error>> {
        let mut errors = Vec::new();

        let address = push_error!(self.address.parse::<Ipv4Addr>().map_err(|e| {
            Error::from(RecordParseError::Ip4Address { input: self.address })
                .with_cause(&e.to_string())
                .with_path("/address")
        }), errors);

        if errors.is_empty() {
            Ok(rdata::A::new(address.unwrap()))
        } else {
            Err(errors)
        }
    }
}

/* --------- AAAA --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Aaaa {
    pub address: String,
}

impl From<rdata::Aaaa> for Aaaa {
    fn from(record_data: rdata::Aaaa) -> Self {
        Aaaa { address: record_data.addr().to_string() }
    }
}

impl Aaaa {
    pub fn parse_record(self) -> Result<rdata::Aaaa, Vec<Error>> {
        let mut errors = Vec::new();

        let address = push_error!(self.address.parse::<Ipv6Addr>().map_err(|e| {
            Error::from(RecordParseError::Ip6Address { input: self.address })
                .with_cause(&e.to_string())
                .with_path("/address")
        }), errors);

        if errors.is_empty() {
            Ok(rdata::Aaaa::new(address.unwrap()))
        } else {
            Err(errors)
        }
    }
}

/* --------- CNAME --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Cname {
    pub target: String,
}

impl<N: ToString> From<rdata::Cname<N>> for Cname {
    fn from(record_data: rdata::Cname<N>) -> Self {
        Cname { target: record_data.cname().to_string() }
    }
}

impl Cname {
    pub fn parse_record(self) -> Result<rdata::Cname<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let cname = parse_name!(self, target, "CNAME", errors);

        if errors.is_empty() {
            Ok(rdata::Cname::new(cname.unwrap()))
        } else {
            Err(errors)
        }

    }
}

/* --------- MX --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Mx {
    pub preference: u16,
    pub mail_exchanger: String,
}

impl<N: ToString> From<rdata::Mx<N>> for Mx {
    fn from(record_data: rdata::Mx<N>) -> Self {
        Mx {
            preference: record_data.preference(),
            mail_exchanger: record_data.exchange().to_string()
        }
    }
}

impl Mx {
    fn parse_record(self) -> Result<rdata::Mx<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let mail_exchanger = parse_name!(self, mail_exchanger, "MX", errors);

        if errors.is_empty() {
            Ok(rdata::Mx::new(self.preference, mail_exchanger.unwrap()))
        } else {
            Err(errors)
        }
    }
}

/* --------- NS --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Ns {
    pub target: String,
}

impl<N: ToString> From<rdata::Ns<N>> for Ns  {
    fn from(record_rdata: rdata::Ns<N>) -> Self {
        Ns {
            target: record_rdata.nsdname().to_string(),
        }
    }
}

impl Ns {
    fn parse_record(self) -> Result<rdata::Ns<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let ns_name = parse_name!(self, target, "NS", errors);

        if errors.is_empty() {
            Ok(rdata::Ns::new(ns_name.unwrap()))
        } else {
            Err(errors)
        }
    }
}

/* --------- PTR --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Ptr {
    pub target: String,
}

impl<N: ToString> From<rdata::Ptr<N>> for Ptr  {
    fn from(record_rdata: rdata::Ptr<N>) -> Self {
        Ptr {
            target: record_rdata.ptrdname().to_string(),
        }
    }
}

impl Ptr {
    fn parse_record(self) -> Result<rdata::Ptr<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let ptr_name = parse_name!(self, target, "PTR", errors);

        if errors.is_empty() {
            Ok(rdata::Ptr::new(ptr_name.unwrap()))
        } else {
            Err(errors)
        }
    }
}

/* --------- SOA --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Soa {
    pub primary_server: String,
    pub maintainer: String,
    pub refresh: u32,
    pub retry: u32,
    pub expire: u32,
    pub minimum: u32,
    pub serial: u32,
}

impl<N: ToString> From<rdata::Soa<N>> for Soa {
    fn from(record_rdata: rdata::Soa<N>) -> Self {
        Soa {
            primary_server: record_rdata.mname().to_string(),
            maintainer: record_rdata.rname().to_string(),
            refresh: record_rdata.refresh().as_secs(),
            retry: record_rdata.retry().as_secs(),
            expire: record_rdata.expire().as_secs(),
            minimum: record_rdata.minimum().as_secs(),
            serial: record_rdata.serial().into(),
        }
    }
}

impl Soa {
    fn parse_record(self) -> Result<rdata::Soa<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let primary_ns = parse_name!(self, primary_server, "SOA", errors);
        let maintainer = parse_name!(self, maintainer, "SOA", errors);

        if errors.is_empty() {
            Ok(rdata::Soa::new(
                primary_ns.unwrap(),
                maintainer.unwrap(),
                self.refresh.into(),
                Ttl::from_secs(self.retry),
                Ttl::from_secs(self.expire),
                Ttl::from_secs(self.minimum),
                Ttl::from_secs(self.serial),
            ))
        } else {
            Err(errors)
        }
    }
}

/* --------- SRV --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Srv {
    pub server: String,
    pub port: u16,
    pub priority: u16,
    pub weight: u16,
}

impl<N: ToString> From<rdata::Srv<N>> for Srv {
    fn from(record_data: rdata::Srv<N>) -> Self {
        Srv {
            server: record_data.target().to_string(),
            priority: record_data.priority(),
            weight: record_data.weight(),
            port: record_data.port(),
        }
    }
}

impl Srv {
    fn parse_record(self) -> Result<rdata::Srv<Name<Vec<u8>>>, Vec<Error>> {
        let mut errors = Vec::new();

        let server = parse_name!(self, server, "SRV", errors);

        if errors.is_empty() {
            Ok(rdata::Srv::new(
                self.priority,
                self.weight,
                self.port,
                server.unwrap(),
            ))
        } else {
            Err(errors)
        }
    }
}


/* --------- TXT --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Txt {
    pub text: String,
}

impl<O: AsRef<[u8]>> From<rdata::Txt<O>> for Txt {
    fn from(record_data: rdata::Txt<O>) -> Self {
        let mut concatenated_text = String::new();
        for text in record_data.iter() {
            for c in text {
                // Escapes '\' and non printable chars
                let c = Symbol::display_from_octet(*c);
                write!(concatenated_text, "{}", c).unwrap();
            }
        }

        Txt {
            text: concatenated_text
        }
    }
}

impl Txt {
    fn parse_record(self) -> Result<rdata::Txt<Vec<u8>>, Vec<Error>> {
        let mut errors = Vec::new();
        let data = append_errors!(validation::parse_txt_data(&self.text), errors, "/text");
        let data = data.and_then(|data| {
            push_error!(rdata::Txt::build_from_slice(&data).map_err(|e| {
                Error::from(RecordParseError::RDataUnknown {
                    input: self.text,
                    field: "text".into(),
                    rtype: "TXT".into(),
                }).with_cause(&e.to_string())
                    .with_path("/text")
            }), errors)
        });



        if errors.is_empty() {
            Ok(data.unwrap())
        } else {
            Err(errors)
        }
    }
}

/* --------- ParsedRData: domain traits impl --------- */

impl<Name, Octs> ParsedRData<Name, Octs> {
    pub fn rtype(&self) -> Rtype {
        match self {
            ParsedRData::A(_) => Rtype::A,
            ParsedRData::Aaaa(_) => Rtype::AAAA,
            ParsedRData::Cname(_) => Rtype::CNAME,
            ParsedRData::Mx(_) => Rtype::MX,
            ParsedRData::Ns(_) => Rtype::NS,
            ParsedRData::Ptr(_) => Rtype::PTR,
            ParsedRData::Soa(_) => Rtype::SOA,
            ParsedRData::Srv(_) => Rtype::SRV,
            ParsedRData::Txt(_) => Rtype::TXT,
        }
    }
}

impl<Name, Octs> RecordData for ParsedRData<Name, Octs> {
    fn rtype(&self) -> Rtype {
        ParsedRData::rtype(self)
    }
}

impl<'a, Octs: Octets + ?Sized> ParseRecordData<'a, Octs> for ParsedRData<ParsedName<Octs::Range<'a>>, Octs::Range<'a>> {
    fn parse_rdata(
        rtype: Rtype,
        parser: &mut Parser<'a, Octs>,
    ) -> Result<Option<Self>, ParseError> {
        let record = match rtype {
            Rtype::A => ParsedRData::A(rdata::A::parse(parser)?),
            Rtype::AAAA => ParsedRData::Aaaa(rdata::Aaaa::parse(parser)?),
            Rtype::CNAME => ParsedRData::Cname(rdata::Cname::parse(parser)?),
            Rtype::MX => ParsedRData::Mx(rdata::Mx::parse(parser)?),
            Rtype::NS => ParsedRData::Ns(rdata::Ns::parse(parser)?),
            Rtype::PTR => ParsedRData::Ptr(rdata::Ptr::parse(parser)?),
            Rtype::SOA => ParsedRData::Soa(rdata::Soa::parse(parser)?),
            Rtype::SRV => ParsedRData::Srv(rdata::Srv::parse(parser)?),
            Rtype::TXT => ParsedRData::Txt(rdata::Txt::parse(parser)?),
            _ => return Ok(None)
        };

        Ok(Some(record))
    }
}

impl<Name: ToName, Octs: AsRef<[u8]>> ComposeRecordData for ParsedRData<Name, Octs> {
    fn rdlen(&self, compress: bool) -> Option<u16> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Aaaa(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Cname(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Mx(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Ns(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Ptr(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Soa(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Srv(record_rdata) => record_rdata.rdlen(compress),
            ParsedRData::Txt(record_rdata) => record_rdata.rdlen(compress),
        }
    }

    fn compose_rdata<Target: Composer + ?Sized>(
        &self,
        target: &mut Target,
    ) -> Result<(), Target::AppendError> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Aaaa(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Cname(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Mx(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Ns(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Ptr(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Soa(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Srv(record_rdata) => record_rdata.compose_rdata(target),
            ParsedRData::Txt(record_rdata) => record_rdata.compose_rdata(target),
        }
    }

    fn compose_canonical_rdata<Target: Composer + ?Sized>(
        &self,
        target: &mut Target,
    ) -> Result<(), Target::AppendError> {
        match self {
            ParsedRData::A(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Aaaa(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Cname(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Mx(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Ns(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Ptr(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Soa(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Srv(record_rdata) => record_rdata.compose_canonical_rdata(target),
            ParsedRData::Txt(record_rdata) => record_rdata.compose_canonical_rdata(target),
        }
    }
}
